/*
 * Copyright 2002 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *  
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <sys/types.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "dag.h"

extern GtkWidget *drawingarea;
extern GtkStyle *red;
extern GtkStyle *yellow;
extern GtkStyle *blue;
extern GtkStyle *green;

/* Graphics options */
static int mode_drawlines = 1;
static int mode_drawnodes = 1;
static int mode_loadtopology = -1;

/* Backing pixmap for drawing area */
static GdkPixmap *pixmap = NULL;
int want_nodes = 50;
int num_iterations = 1;
static int stop_iterate;
static int pixmap_width;
static int pixmap_height;

/* Draw a rectangle on the screen */
static void
draw_brush (GtkWidget *widget, gdouble x, gdouble y)
{
	GdkRectangle update_rect;

	update_rect.x = x - 5;
	update_rect.y = y - 5;
	update_rect.width = 10;
	update_rect.height = 10;

	gdk_draw_rectangle (pixmap,
	    widget->style->black_gc,
	    TRUE,
	    update_rect.x, update_rect.y,
	    update_rect.width, update_rect.height);

	gtk_widget_draw (widget, &update_rect);
}

void
on_draw_point(GtkWidget *widget, GtkStyle *style, int x, int y, int size)
{
	GdkRectangle update_rect;

	update_rect.x = x - size/2;
	update_rect.y = y - size/2;
	update_rect.width = size;
	update_rect.height = size;

	gdk_draw_arc(pixmap, style->fg_gc[GTK_STATE_NORMAL], 1,
	    update_rect.x, update_rect.y,
	    update_rect.width, update_rect.height,
	    0, 360 * 64);
}

void
on_draw_line(GtkWidget *widget, GdkGC *gc, int x1, int y1, int x2, int y2)
{
	GdkRectangle update_rect;
	int xs, ys, xe, ye;

	if (x1 < x2) {
		xs = x1;
		xe = x2;
	} else {
		xs = x2;
		xe = x1;
	}

	if (y1 < y2) {
		ys = y1;
		ye = y2;
	} else {
		ys = y2;
		ye = y1;
	}

	update_rect.x = xs;
	update_rect.y = ys;
	update_rect.width = xe - xs + 1;
	update_rect.height = ye - ys + 1;

	gdk_draw_line(pixmap, gc, x1, y1, x2, y2);
}

void
node_type(GtkWidget *widget, struct dag_node *node,
    int *psize, GtkStyle **pstyle)
{
	int size;
	GtkStyle *style;

	style = widget->style;
	size = 2;
	if (node->size > 1) {
		style = blue;
		size = 4;
	}
	if (node->size > 3) {
		style = green;
		size = 6;
	}
	if (node->size > 8) {
		style = red;
		size = 8;
	}
	if (node->size > 10) {
		int nr = node->size - 10;

		size += (nr > 10 ? 10 : nr) / 3;
		if (nr > 10) {
			nr -= 10;
			size += (nr > 10 ? 10 : nr) / 3;
		}
	}
	size += 2;

	if (psize != NULL)
		*psize = size;
	if (pstyle != NULL)
		*pstyle = style;
}

void
draw_node(GtkWidget *widget, struct dag_node *node)
{
	GtkStyle *style;
	struct dag_edge *edge;
	struct dag_node *dest;
	int count = 0, size;

	node_type(widget, node, &size, &style);

	if (mode_drawlines) {
		SPLAY_FOREACH(edge, edgetree, &node->edgeroot) {
			GtkStyle *linesty;
			count++;
			dest = edge->destination;

			if (edge->undirected && dest < node) 
				continue;

			linesty = style;
			if (dest->size < node->size)
				node_type(widget, dest, NULL, &linesty);

			on_draw_line(widget, linesty->fg_gc[GTK_STATE_NORMAL],
			    node->p_x, node->p_y, dest->p_x, dest->p_y);
		}
	}

	if (mode_drawnodes)
		on_draw_point(widget, style, node->p_x, node->p_y, size);
}

int
dag_cb(struct dag_node *node, void *widget)
{
	draw_node(GTK_WIDGET(widget), node);

	return (0);
}

void
redraw(GtkWidget *widget)
{
	static int empty;
	GdkRectangle update_rect;

	if (empty && !mode_drawnodes && !mode_drawlines)
		return;

	gdk_draw_rectangle (pixmap,
	    widget->style->white_gc,
	    TRUE,
	    0, 0,
	    pixmap_width, pixmap_height);

	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = widget->allocation.width;
	update_rect.height = widget->allocation.height;

	if (mode_drawnodes || mode_drawlines) {
		dag_walk_all(dag_cb, widget);
		empty = 0;
	} else
		empty = 1;

	gtk_widget_draw (widget, &update_rect);
}

gboolean
on_drawingarea1_configure_event        (GtkWidget       *widget,
                                        GdkEventConfigure *event,
                                        gpointer         user_data)
{
	int width, height;

	width = widget->allocation.width;
	height = widget->allocation.height;

	if (pixmap)
		gdk_pixmap_unref(pixmap);

	if (width < pixmap_width || height < pixmap_height) {
		gtk_drawing_area_size(GTK_DRAWING_AREA(widget),
		    pixmap_width, pixmap_height);
		return FALSE;
	}

	pixmap_width = width;
	pixmap_height = height;

	pixmap = gdk_pixmap_new(widget->window, pixmap_width, pixmap_height,
	    -1);

	dag_update_max(pixmap_width, pixmap_height);

	redraw(widget);

	return TRUE;
}


gboolean
on_drawingarea1_expose_event           (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	gdk_draw_pixmap(widget->window,
	    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
	    pixmap,
	    event->area.x, event->area.y,
	    event->area.x, event->area.y,
	    event->area.width, event->area.height);

	return FALSE;
}

int
dag_brush_cb(struct dag_node *node, void *arg)
{
	struct dag_point point;
	int offx, offy;
	
	offx = (arc4random() % 40) - 20;
	offy = (arc4random() % 40) - 20;
	
	point.x = node->p_x + offx;
	point.y = node->p_y + offy;

	dag_move_node(node, &point);

	redraw(drawingarea);

	return (0);
}


gboolean
on_drawingarea1_button_press_event     (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
	if (event->button != 1 || pixmap == NULL)
		return (TRUE);

	draw_brush (widget, event->x, event->y);

	dag_walk_region(dag_brush_cb,
	    event->x - 20, event->y - 20,
	    event->x + 20, event->y + 20, NULL);

	return TRUE;
}


gboolean
on_drawingarea1_motion_notify_event    (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
	GdkModifierType state;
	int x, y;

	if (event->is_hint)
		gdk_window_get_pointer (event->window, &x, &y, &state);
	else {
		x = event->x;
		y = event->y;
		state = event->state;
	}
   
	if (state & GDK_BUTTON1_MASK && pixmap != NULL)
		draw_brush (widget, x, y);
 
	return TRUE;
}


void
on_entry1_changed                      (GtkEditable     *editable,
                                        gpointer         user_data)
{
	char *text, *end;
	int nodes;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	if (!strlen(text))
		return;

	nodes = strtoul(text, &end, 10);
	if (*end == '\0')
		want_nodes = nodes;
}

void
on_new_button_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	struct dag_node *node, *second;
	int i;

	for (i = 0; i < want_nodes; i++) {
		int x, y;

		do {
			x = 3 + (arc4random() % (pixmap_width - 6));
			y = 3 + (arc4random() % (pixmap_height - 6));

			node = dag_new_node(x, y);
		} while (node == NULL);

		do {
			second = dag_random_node();
		} while (second != NULL && second == node);

		dag_insert_node(node);

		if (second == NULL)
			continue;

		dag_connect(node, second, 1);

		if (i % 10 != 9)
			continue;

		node = dag_random_node();
		second = dag_random_node();
		if (node != second && !dag_isconnected(node, second))
			dag_connect(node, second, 1);
	}

	redraw(drawingarea);
}


void
on_button2_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
	gtk_main_quit();
}


void
on_button3_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
	char line[40];
	struct timeval tv_start, tv_end;
	GtkWidget *label_time, *label_energy;
	double oldenergy, stepsize, energy;
	int i, nochange;

	label_time = lookup_widget(GTK_WIDGET(button), "time");
	label_energy = lookup_widget(GTK_WIDGET(button), "energy");

	gtk_widget_set_sensitive(GTK_WIDGET(button), 0);

	dag_update_max(pixmap_width, pixmap_height);

	stepsize = dag_set_stepsize(0);

	stop_iterate = 0;
	oldenergy = 0;
	nochange = 0;
	for (i = 0; (num_iterations == 0 || i < num_iterations)
	    && !stop_iterate; i++) {
		gettimeofday(&tv_start, NULL);
		dag_hash();
		energy = dag_compute_forces();
		dag_apply_forces();
		dag_compute_forces();
		dag_find_minimum();
		gettimeofday(&tv_end, NULL);

		timersub(&tv_end, &tv_start, &tv_end);

		/* Do not update every time */
		if (tv_end.tv_sec || (i % 5) == 0) {
			snprintf(line, sizeof(line), "%2d:%02d.%02d",
			    (int)(tv_end.tv_sec / 60),
			    (int)(tv_end.tv_sec % 60),
			    (int)(tv_end.tv_usec / 10000));
			gtk_label_set_text(GTK_LABEL(label_time), line);
			snprintf(line, sizeof(line), "%12.2f", energy);

			gtk_label_set_text(GTK_LABEL(label_energy), line);
		}
		
		if ((i % 5) == 0)
			redraw(drawingarea);

		while (gtk_events_pending())
			if (gtk_main_iteration_do(0)) {
				stop_iterate = 1;
				break;
			}

		if (oldenergy) {
			if (1.1 * oldenergy < energy) {
				nochange = 0;
				stepsize /= 2;
				dag_set_stepsize(stepsize);
			} else if (nochange > 20 || oldenergy > 1.5*energy) {
				nochange = 0;
				stepsize *= 1.1;
				dag_set_stepsize(stepsize);
			} else if (oldenergy > energy)
				nochange++;
		}
		oldenergy = energy;
	}

	gtk_widget_set_sensitive(GTK_WIDGET(button), 1);
}


void
on_nriterations_changed                (GtkEditable     *editable,
                                        gpointer         user_data)
{
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	num_iterations = atoi(text);
}

void
on_stop_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
	stop_iterate = 1;
}


void
on_repulsedist_changed                 (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repdist, attrdist, repcenter, attrcenter;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_params(&repdist, &attrdist, &repcenter, &attrcenter);
	repdist = atof(text);
	dag_set_params(repdist, attrdist, repcenter, attrcenter);
}


void
on_repulsecenter_changed               (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repdist, attrdist, repcenter, attrcenter;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_params(&repdist, &attrdist, &repcenter, &attrcenter);
	repcenter = atof(text);
	dag_set_params(repdist, attrdist, repcenter, attrcenter);
}


void
on_attractdist_changed                 (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repdist, attrdist, repcenter, attrcenter;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_params(&repdist, &attrdist, &repcenter, &attrcenter);
	attrdist = atof(text);
	dag_set_params(repdist, attrdist, repcenter, attrcenter);
}


void
on_attractcenter_changed               (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repdist, attrdist, repcenter, attrcenter;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_params(&repdist, &attrdist, &repcenter, &attrcenter);
	attrcenter = atof(text);
	dag_set_params(repdist, attrdist, repcenter, attrcenter);
}


void
on_repulseconst_changed                (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repulse, attract;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_physics(&repulse, &attract);
	repulse = atof(text);
	dag_set_physics(repulse, attract);
}


void
on_attractconst_changed                (GtkEditable     *editable,
                                        gpointer         user_data)
{
	double repulse, attract;
	char *text;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	dag_get_physics(&repulse, &attract);
	attract = atof(text);
	dag_set_physics(repulse, attract);
}


void
on_file1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_open1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	extern GtkWidget *loadtopology;

	if (mode_loadtopology != -1)
		return;

	mode_loadtopology = 1;

	gtk_window_set_title(GTK_WINDOW(loadtopology), "Load a topology");
	gtk_widget_show(loadtopology);
}


void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	gtk_main_quit();
}


void
on_cancel_button1_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	extern GtkWidget *loadtopology;

	gtk_widget_hide(loadtopology);

	mode_loadtopology = -1;
}


void
on_ok_button1_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	extern GtkWidget *loadtopology;
	char *name;

	/* Update canvas size */
	dag_update_max(pixmap_width, pixmap_height);

	name = gtk_file_selection_get_filename(GTK_FILE_SELECTION(loadtopology));
	
	if (mode_loadtopology == 1) {
		dag_load_topology(name);

		dag_set_max(pixmap_width, pixmap_height);

		redraw(drawingarea);
	} else if (mode_loadtopology == 0) {
		struct stat sb;

		if (stat(name, &sb) != -1) {
			extern GtkWidget *warningdialog;

			gtk_widget_hide(loadtopology);
			gtk_widget_show(warningdialog);
			return;
		}

		dag_save_topology(name);
	}

	gtk_widget_hide(loadtopology);

	mode_loadtopology = -1;
}


void
on_about1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	extern GtkWidget *aboutwindow;

	gtk_widget_show(aboutwindow);
}


void
on_about_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
	extern GtkWidget *aboutwindow;

	gtk_widget_hide(aboutwindow);
}

void
on_drawlines_toggled                   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	mode_drawlines = mode_drawlines ? 0 : 1;

	redraw(drawingarea);
}


void
on_drawnodes_toggled                   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	mode_drawnodes = mode_drawnodes ? 0 : 1;

	redraw(drawingarea);
}


void
on_save_as1_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	extern GtkWidget *loadtopology;

	if (mode_loadtopology != -1)
		return;

	mode_loadtopology = 0;

	gtk_window_set_title(GTK_WINDOW(loadtopology), "Save curret topology");
	gtk_widget_show(loadtopology);
}


void
on_waning_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
	extern GtkWidget *loadtopology;
	extern GtkWidget *warningdialog;
	char *name;

	if (mode_loadtopology == 0) {
		name = gtk_file_selection_get_filename(GTK_FILE_SELECTION(loadtopology));
		dag_save_topology(name);
	}

	gtk_widget_hide(warningdialog);

	mode_loadtopology = -1;
}


void
on_warning_cancel_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	extern GtkWidget *warningdialog;

	gtk_widget_hide(warningdialog);

	mode_loadtopology = -1;
}

