/*
 * Copyright 2002 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/queue.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <assert.h>

#include "dag.h"

static struct dag_point adjust;
static struct dag_point offset;

char *
fgetl(char *s, int size, FILE *stream)
{
        char *res, *pos;
        int c;

        if ((res = fgets(s, size, stream))) {
                if (!*res) return res;

                pos = res + strlen(res) - 1;
                if (*pos == '\n') {
                        *pos = 0;
                        if (pos > res)
				if (*--pos == '\r') *pos = 0;
                } else
			if ((c = getc(stream)) == '\n') {
				if (*pos == '\r') *pos = 0;
			} else
				while (c != EOF && c != '\n')
					c = getc(stream);
        }

        return (res);
}

int
dag_lt_adjust_cb(struct dag_node *node, void *arg)
{
	struct dag_point *point = arg;

	node->p_x /= point->x;
	node->p_y /= point->y;

	node->p_x += 5;
	node->p_y += 5;

	return (0);
}

#define WHITESPACE " \t"

int
dag_load_topology(char *name)
{
	char line[120], *p;
	FILE *fp;
	int width, height;	/* Canvas */
	int maxx, maxy;
	struct dag_node *node;
	struct dag_point point;
	int id, lastid, nrconnects;

	if ((fp = fopen(name, "r")) == 0) {
		warn("%s: fopen(%s)", __func__, name);
		return (-1);
	}

	dag_get_size(&width, &height);
	width -= 10;
	height -= 10;

	if (fgetl(line, sizeof(line), fp) == NULL) {
		warn("%s: early end of file %s", __func__, name);
		goto closerr;
	}

	lastid = strtoul(line, &p, 10);
	if (p == NULL)
		goto error;
	p += strspn(p, WHITESPACE);
	nrconnects = strtoul(p, &p, 10);

	if (nrconnects <= lastid) {
		warn("%s: number of connections too small", __func__);
		goto closerr;
	}

	maxx = maxy = 0;

	for (id = 0; id < lastid; id++) {
		int rid, x, y;

		if (fgetl(line, sizeof(line), fp) == NULL) {
			warn("%s: early end of file %s", __func__, name);
			goto closerr;
		}


		p = line + strspn(line, WHITESPACE);
		rid = strtoul(p, &p, 10);
		if (p == NULL)
			goto error;

		p += strspn(p, WHITESPACE);
		x = strtoul(p, &p, 10);
		if (p == NULL)
			goto error;

		p += strspn(p, WHITESPACE);
		y = strtoul(p, &p, 10);
		if (p == NULL)
			goto error;

		if (rid != id) {
			warn("%s: expected id %d, but got %d", __func__,
			    id, rid);
			goto closerr;
		}

		node = dag_new_node(x, y);
		if (node == NULL || node->id != id ||
		    node->p_x != x || node->p_y != y) {
			warnx("%s: could not generate node %d -> %d,%d",
			    __func__, id, x, y);
			goto closerr;
		}

		dag_insert_node(node);

		if (x > maxx)
			maxx = x;
		if (y > maxy)
			maxy = y;
	}

	point.x = maxx/(double)width;
	point.y = maxy/(double)height;

	dag_walk_all(dag_lt_adjust_cb, &point);

	for (id = 0; id < nrconnects; id++) {
		int one, two;
		struct dag_node *from, *dest;

		if (fgetl(line, sizeof(line), fp) == NULL) {
			warn("%s: early end of file %s", __func__, name);
			goto closerr;
		}


		p = line + strspn(line, WHITESPACE);
		one = strtoul(p, &p, 10);
		if (p == NULL)
			goto error;

		p += strspn(p, WHITESPACE);
		two = strtoul(p, &p, 10);
		if (p == NULL)
			goto error;

		from = dag_node_findid(one);
		dest = dag_node_findid(two);

		if (from == NULL || dest == NULL) {
			warn("%s: can not find nodes %d(%p) -> %d(%p)",
			    __func__, one, from, two, dest);
			goto closerr;
		}

		dag_connect(from, dest, 1);
	}

	return (0);

 error:
	warnx("%s: parsing error", __func__);

 closerr:
	fclose(fp);
	return (-1);
}

int
dag_count_cb(struct dag_node *node, void *arg)
{
	int *count = arg;

	(*count)++;

	if (node->p_x < offset.x)
		offset.x = node->p_x;
	if (node->p_y < offset.y)
		offset.y = node->p_y;

	return (0);
}

int
dag_count_edges_cb(struct dag_node *node, void *arg)
{
	int *pcount = arg, count = *pcount;;

	count += (int)node->size;

	*pcount = count;

	return (0);
}

int
dag_write_nodes_cb(struct dag_node *node, void *arg)
{
	FILE *fp = arg;

	fprintf(fp, "%d\t%d\t%d\n",
	    node->id,
	    (int)((node->p_x - offset.x) * 10000 / adjust.x),
	    (int)((node->p_y - offset.y) * 10000 / adjust.y));

	return (0);
}

int
dag_write_edges_cb(struct dag_node *node, void *arg)
{
	struct dag_node *dest, one, two;
	struct dag_edge *tmp;
	FILE *fp = arg;

	SPLAY_FOREACH(tmp, edgetree, &node->edgeroot) {
		dest = tmp->destination;
		if (node->id > dest->id)
			continue;

		one = *node;
		two = *dest;

		one.p_x -= offset.x;
		one.p_y -= offset.y;
		two.p_x -= offset.x;
		two.p_y -= offset.y;

		one.p_x *= 10000/adjust.x;
		one.p_y *= 10000/adjust.y;
		two.p_x *= 10000/adjust.x;
		two.p_y *= 10000/adjust.y;

		fprintf(fp, "%d\t%d\t%d\n",
		    node->id, dest->id,
		    (int)dag_distance(&one, &two));
	}

	return (0);
}

int
dag_save_topology(char *name)
{
	char tmpname[30];
	FILE *fp;
	int fd, count = 0, edge = 0;
	int width, height;

	strlcpy(tmpname, "/tmp/netlayout.XXXXXX", sizeof(tmpname));
	if ((fd = mkstemp(tmpname)) == -1) {
		warn("%s: can not open tmp file: %s", __func__, tmpname);
		return (-1);
	}

	if ((fp = fdopen(fd, "w")) == NULL) {
		warn("%s: fopen(%s)", __func__, name);
		return (-1);
	}

	dag_get_size(&width, &height);
	offset.x = adjust.x = width;
	offset.y = adjust.y = height;
	

	dag_walk_all(dag_count_cb, &count);
	dag_walk_all(dag_count_edges_cb, &edge);

	edge /= 2;

	fprintf(fp, "%d %d\n", count, edge);

	adjust.x -= offset.x;
	adjust.y -= offset.y;

	dag_walk_all(dag_write_nodes_cb, fp);
	dag_walk_all(dag_write_edges_cb, fp);

	fclose(fp);

	if (rename(tmpname, name) == -1)
		warn("%s: can not rename %s to %s", __func__, tmpname, name);

	return (0);
}
